#include <ReWindow/types/Window.h>
#include <ReWindow/Logger.h>
#include <GL/glx.h>
#include <dlfcn.h>

using namespace ReWindow;

void* glx_lib = nullptr;
void* opengl_lib = nullptr;

typedef const GLubyte* (*glGetString_t)(GLenum name);
glGetString_t func_glGetString = nullptr;

typedef Bool (*glXMakeCurrent_t)(Display *dpy, GLXDrawable drawable, GLXContext ctx);
glXMakeCurrent_t func_glXMakeCurrent = nullptr;

typedef void* (*glXGetProcAddressARB_t)(const GLubyte *procname);
glXGetProcAddressARB_t func_glXGetProcAddressARB = nullptr;

typedef void (*glXSwapBuffers_t)(Display *dpy, GLXDrawable drawable);
glXSwapBuffers_t func_glXSwapBuffers = nullptr;

typedef GLXContext (*glXCreateContext_t)(Display *dpy, XVisualInfo *vis, GLXContext shareList, Bool direct);
glXCreateContext_t func_glXCreateContext = nullptr;

typedef GLXFBConfig *(*glXChooseFBConfig_t)(Display *dpy, int screen, const int *attribList, int *nItems);
glXChooseFBConfig_t func_glXChooseFBConfig = nullptr;

typedef XVisualInfo* (*glXGetVisualFromFBConfig_t)(Display *dpy, GLXFBConfig config);
glXGetVisualFromFBConfig_t func_glXGetVisualFromFBConfig = nullptr;

typedef XVisualInfo* (*glXChooseVisual_t)(Display *dpy, int screen, const int *attribList);
glXChooseVisual_t func_glXChooseVisual = nullptr;

typedef GLXContext (*glXCreateContextAttribsARBProc)(Display*, GLXFBConfig, GLXContext, Bool, const int*);
GLXContext gl_context;

void OpenGLWindow::SwapBuffers() {
    func_glXSwapBuffers(xVars.display,xVars.window);
}

void OpenGLWindow::SetVsyncEnabled(bool state) {
    PFNGLXSWAPINTERVALEXTPROC glXSwapIntervalEXT;
    glXSwapIntervalEXT = (PFNGLXSWAPINTERVALEXTPROC) func_glXGetProcAddressARB((const GLubyte *) "glXSwapIntervalEXT");
    glXSwapIntervalEXT(xVars.display, xVars.window, state);
}

std::string OpenGLWindow::GetGraphicsDriverVendor() {
    return std::string((const char*) func_glGetString(GL_VENDOR));
}

void OpenGLWindow::Open() {
    xVars.display = XOpenDisplay(nullptr);
    xVars.defaultScreen = DefaultScreen(xVars.display);

    xVars.xSetWindowAttributes.border_pixel = BlackPixel(xVars.display, xVars.defaultScreen);
    xVars.xSetWindowAttributes.background_pixel = BlackPixel(xVars.display, xVars.defaultScreen);
    xVars.xSetWindowAttributes.override_redirect = True;
    xVars.xSetWindowAttributes.event_mask = ExposureMask;

    auto glXCreateContextAttribsARB = (glXCreateContextAttribsARBProc)func_glXGetProcAddressARB((const GLubyte*)"glXCreateContextAttribsARB");
    XVisualInfo* vi = nullptr;

    // Fallback to the old way if you somehow don't have this.
    if (!glXCreateContextAttribsARB) {
        GLint glAttributes[] = {GLX_RGBA, GLX_DEPTH_SIZE, 24, GLX_DOUBLEBUFFER, None};
        xVars.visual = func_glXChooseVisual(xVars.display, xVars.defaultScreen, glAttributes);
        gl_context = func_glXCreateContext(xVars.display, xVars.visual, nullptr, GL_TRUE);
        xVars.window = XCreateWindow(xVars.display, RootWindow(xVars.display, xVars.defaultScreen), 0, 0, width, height, 0, xVars.visual->depth,
                                     InputOutput, xVars.visual->visual, CWBackPixel | CWColormap | CWBorderPixel | NoEventMask, &xVars.xSetWindowAttributes);
        ReWindow::Logger::Debug("Created OpenGL Context with glXCreateContext.");
    }
    else {
        static int visual_attributes[]
                {
                        GLX_X_RENDERABLE, True, GLX_DRAWABLE_TYPE, GLX_WINDOW_BIT,
                        GLX_RENDER_TYPE, GLX_RGBA_BIT, GLX_X_VISUAL_TYPE, GLX_TRUE_COLOR,
                        GLX_RED_SIZE, 8, GLX_GREEN_SIZE, 8, GLX_BLUE_SIZE, 8,
                        GLX_ALPHA_SIZE, 8, GLX_DEPTH_SIZE, 24, GLX_STENCIL_SIZE, 8,
                        GLX_DOUBLEBUFFER, True, None
                };

        int fb_count;
        GLXFBConfig* fb_configurations = func_glXChooseFBConfig(xVars.display, xVars.defaultScreen, visual_attributes, &fb_count);
        if (!fb_configurations || fb_count == 0)
            throw std::runtime_error("Couldn't get framebuffer configuration.");

        GLXFBConfig best_fbc = fb_configurations[0];
        vi = func_glXGetVisualFromFBConfig(xVars.display, best_fbc);

        if (!vi)
            throw std::runtime_error("Couldn't visual info from framebuffer configuration.");

        xVars.xSetWindowAttributes.colormap = XCreateColormap(xVars.display, RootWindow(xVars.display, vi->screen), vi->visual, AllocNone);
        xVars.window = XCreateWindow(xVars.display, RootWindow(xVars.display, vi->screen), 0, 0, width, height, 0, vi->depth, InputOutput,
                                     vi->visual, CWBackPixel | CWColormap | CWBorderPixel, &xVars.xSetWindowAttributes);

        // TODO allow the user to specify what OpenGL version they want.
        int context_attributes[] { GLX_CONTEXT_MAJOR_VERSION_ARB, gl_major, GLX_CONTEXT_MINOR_VERSION_ARB, gl_minor, None };
        gl_context = glXCreateContextAttribsARB(xVars.display, best_fbc, nullptr, True, context_attributes);
        XFree(fb_configurations);
        ReWindow::Logger::Debug("Created OpenGL Context with glXCreateContextAttribsARB");
    }

    if (!gl_context)
        throw std::runtime_error("Couldn't create the OpenGL context.");
    if (!func_glXMakeCurrent(xVars.display, xVars.window, gl_context))
        throw std::runtime_error("Couldn't change OpenGL context to current.");

    if (vi)
        XFree(vi);

    XSelectInput(xVars.display, xVars.window, ExposureMask | KeyPressMask | KeyReleaseMask | ButtonPressMask | ButtonReleaseMask | PointerMotionMask |
    PointerMotionHintMask | FocusChangeMask | StructureNotifyMask | SubstructureRedirectMask | SubstructureNotifyMask | CWColormap );

    XMapWindow(xVars.display, xVars.window);
    XStoreName(xVars.display, xVars.window, title.c_str());

    xVars.windowTypeAtom = XInternAtom(xVars.display, "_NET_WM_WINDOW_TYPE", False);
    xVars.windowTypeUtilityAtom = XInternAtom(xVars.display, "_NET_WM_WINDOW_TYPE_UTILITY", False);
    XChangeProperty(xVars.display, xVars.window, xVars.windowTypeAtom, XA_ATOM, 32, PropModeReplace, (unsigned char*) &xVars.windowTypeUtilityAtom, 1);

    xVars.wmDeleteWindow = XInternAtom(xVars.display, "WM_DELETE_WINDOW", False);
    XSetWMProtocols(xVars.display, xVars.window, &xVars.wmDeleteWindow, 1);

    XGetWindowAttributes(xVars.display, xVars.window, &xVars.windowAttributes);
    render_area_position = { xVars.windowAttributes.x, xVars.windowAttributes.y };

    open = true;
    processOnOpen();
}

OpenGLWindow::OpenGLWindow(const std::string& title, int width, int height, uint8_t gl_major, uint8_t gl_minor)
: gl_major(gl_major), gl_minor(gl_minor), RWindow(title, width, height) {

    // Try this first so that in the event we have more than one window we don't double-load the library.
    glx_lib = dlopen("libGLX.so", RTLD_NOLOAD);
    opengl_lib = dlopen("libOpenGL.so", RTLD_NOLOAD);

    if (!glx_lib)
        glx_lib = dlopen("libGLX.so", RTLD_LAZY);
    if (!opengl_lib)
        opengl_lib = dlopen("libOpenGL.so", RTLD_LAZY);

    if (!opengl_lib)
        throw std::runtime_error("libOpenGL.so couldn't be found in your LD_LIBRARY_PATH.");
    if (!glx_lib)
        throw std::runtime_error("libGLX.so couldn't be found in your LD_LIBRARY_PATH.");

    func_glGetString = (glGetString_t) dlsym(opengl_lib, "glGetString");
    func_glXMakeCurrent = (glXMakeCurrent_t) dlsym(glx_lib, "glXMakeCurrent");
    func_glXGetProcAddressARB = (glXGetProcAddressARB_t) dlsym(glx_lib, "glXGetProcAddressARB");
    func_glXSwapBuffers = (glXSwapBuffers_t) dlsym(glx_lib, "glXSwapBuffers");
    func_glXCreateContext = (glXCreateContext_t) dlsym(glx_lib, "glXCreateContext");
    func_glXChooseFBConfig = (glXChooseFBConfig_t) dlsym(glx_lib, "glXChooseFBConfig");
    func_glXGetVisualFromFBConfig = (glXGetVisualFromFBConfig_t) dlsym(glx_lib, "glXGetVisualFromFBConfig");
    func_glXChooseVisual = (glXChooseVisual_t) dlsym(glx_lib, "glXChooseVisual");
}
